#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/serial.h>
#include <string.h>
#include <asm-generic/ioctls.h>
#include <unistd.h>
#include <sys/io.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/ioctl.h>

#define ACCESS_ONCE(x, y)                    \
	{                                    \
		volatile char *tmp_xxx = &x; \
		*tmp_xxx = y;                \
		usleep(10);                  \
	}

#define ACCESS_GET(x)                        \
	({                                   \
		volatile char *tmp_xxx = &x; \
		*tmp_xxx;                    \
	})

#define F81504_CONFIG_SYSFS_IOPORT "/sys/class/tty/%s/port"
#define F81504_CONFIG_SYSFS_MEM_BASE "/sys/class/tty/%s/iomem_base"
#define F81504_SERIAL_SYSFS_IOPORT "/proc/tty/driver/serial"
#define F81504_SYS_PCI_SYSFS "/sys/bus/pci/devices"

#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof(arr[0]))
#define BIT(x) (1ULL << (x))

#define VERSION "v20241025"

static volatile unsigned char *can_mmio;

static int m_mem_fd = -1;
static unsigned char m_pci_path[128];
static unsigned char m_pci_conf_path[128];

const static char *m_cStringArray_f81439[] = {
	"F81439(M0/M1/M2)",
	"\t0: 000 - RS422",
	"\t1: 001 - RS232",
	"\t2: 010 - RS485 Low Active",
	"\t3: 011 - RS485 High Active",
	"\t4: 100 - RS422 with termination",
	"\t5: 101 - Reserved",
	"\t6: 110 - RS485 Low Active with termination",
	"\t7: 111 - Shutdown",
	NULL,
};

const static int m_accpet_mode_f81439[] = { 0, 1, 2, 3, 4, 5, 6, 7, -1 };

const static char *m_cStringArray_f81435[] = {
	"F81435(M0/M1/M2)",
	"\t0: 000 - RS422 High Active",
	"\t1: 001 - RS232",
	"\t2: 010 - RS485 High Active",
	"\t3: 011 - ccTALK",
	"\t4: 100 - RS422 Low Active",
	"\t5: 101 - RS485 Low Active (Auto Sensing)",
	"\t6: 110 - RS485 Low Active",
	"\t7: 111 - RS422 Low Active (Auto Sensing)",
	NULL,
};

const static int m_accpet_mode_f81435[] = { 0, 1, 2, 3, 4, 5, 6, 7, -1 };

const static char *m_cStringArray_f81433[] = {
	"F81433(M0/M1/R_DIS)",
#if 0
	"\t1: 001 - RS485 Low Active with termination",
	"\t2: 010 - RS422 with termination",
	"\t4: 100 - RS232",
	"\t5: 101 - RS485 Low Active",
	"\t6: 110 - RS422",
	"\t7: 111 - Shutdown",
#else
	"\t1: 001 - RS232",
	"\t2: 010 - RS422 with termination",
	"\t3: 011 - RS422",
	"\t4: 100 - RS485 Low Active with termination (Auto Sensing)",
	"\t5: 101 - RS485 Low Active (Auto Sensing)",
	"\t7: 111 - Shutdown",
#endif
	NULL,
};

const static int m_accpet_mode_f81433[] = { 1, 2, 3, 4, 5, 7, -1 };

typedef enum {
	transceiver_type_f81439,
	transceiver_type_f81435,
	transceiver_type_f81433,
	transceiver_type_invalid,
} transceiver_type;

const static char **m_cStringArray[transceiver_type_invalid] = {
	[transceiver_type_f81439] = m_cStringArray_f81439,
	[transceiver_type_f81435] = m_cStringArray_f81435,
	[transceiver_type_f81433] = m_cStringArray_f81433,
};

const static int *m_nAcceptArray[transceiver_type_invalid] = {
	[transceiver_type_f81439] = m_accpet_mode_f81439,
	[transceiver_type_f81435] = m_accpet_mode_f81435,
	[transceiver_type_f81433] = m_accpet_mode_f81433,
};

const static char *m_nNameArray[transceiver_type_invalid] = {
	[transceiver_type_f81439] = "F81439",
	[transceiver_type_f81435] = "F81435",
	[transceiver_type_f81433] = "F81433",
};

static int read_pci(int addr, void *data, int len)
{
	FILE *file;

	file = fopen(m_pci_conf_path, "r");
	if (!file)
		return -1;

	fseek(file, addr, SEEK_SET);
	len = fread(data, len, 1, file);
	if (len != len) {
		fclose(file);
		return -1;
	}

	fclose(file);

	return 0;
}

static int read_pci_byte(int addr, void *data)
{
	FILE *file;
	int len;

	file = fopen(m_pci_conf_path, "r");
	if (!file)
		return -1;

	fseek(file, addr, SEEK_SET);
	len = fread(data, 1, 1, file);
	if (len != 1) {
		fclose(file);
		return -1;
	}

	fclose(file);

	return 0;
}

static int write_pci_byte(int addr, unsigned char data)
{
	FILE *file;
	int len;

	file = fopen(m_pci_conf_path, "wb");
	if (!file)
		return -1;

	fseek(file, addr, SEEK_SET);
	len = fwrite(&data, 1, 1, file);
	if (len != 1) {
		fclose(file);
		return -1;
	}

	fclose(file);

	return 0;
}

static int write_pci_mask_byte(int addr, unsigned char mask,
			       unsigned char data)
{
	int status;
	unsigned char tmp;

	status = read_pci_byte(addr, &tmp);
	if (status)
		return -1;

	tmp &= ~mask;
	tmp |= (mask & data);

	status = write_pci_byte(addr, tmp);
	if (status)
		return -1;

	return 0;
}

static void usage(char *argv[])
{
	const char **st;
	int i, j;

	fprintf(stderr, "Fintek F81504 set gpio tools\n");
	fprintf(stderr, "Set mode\n");
	fprintf(stderr, "\t%s <port> <mode>\n", argv[0]);
	fprintf(stderr, "Query mode\n");
	fprintf(stderr, "\t%s <port>\n", argv[0]);
	fprintf(stderr, "\nUart Mode table:\n");

	for (i = 0; i < transceiver_type_invalid; ++i) {
		st = m_cStringArray[i];

		for (j = 0; st[j]; ++j)
			fprintf(stderr, "%s\n", st[j]);

		fprintf(stderr, "\n");
	}

	fprintf(stderr, "\n");
}

int is_regular_file(const char *path)
{
	struct stat statbuf;

	// Check if path exists
	if (access(path, F_OK) != 0)
		return -1;

	// Get file status
	if (stat(path, &statbuf) == 0 && S_ISREG(statbuf.st_mode))
		return 0; // It's a directory

	return -1; // Not a directory
}

static int get_res_with_pcie(unsigned int res)
{
	struct dirent **namelist;
	unsigned int pid, vid, bar0, tmp;
	int n, i, j, r, selected = 0;

	n = scandir(F81504_SYS_PCI_SYSFS, &namelist, 0, alphasort);
	if (n < 0)
		return -1;

	for (i = 0; i < n; i++) {
		sprintf(m_pci_path, "%s/%s", F81504_SYS_PCI_SYSFS,
			namelist[i]->d_name);

		sprintf(m_pci_conf_path, "%s/%s/config", F81504_SYS_PCI_SYSFS,
			namelist[i]->d_name);

		if (is_regular_file(m_pci_conf_path))
			continue;

		r = read_pci(0x00, &vid, 2);
		if (r)
			continue;

		r = read_pci(0x02, &pid, 2);
		if (r)
			continue;

		if (vid != 0x1c29)
			continue;

		if ((pid & 0xff00) == 0x1300) {
			r = read_pci(0x10, &bar0, 4);
			if (r)
				continue;

			if (res < bar0) {
				//printf("%s: %d\n", __func__, __LINE__);
				continue;
			}

			if (res - bar0 >= 8 * 12) {
				//printf("%s: %d %x %x\n", __func__, __LINE__,
				//       bar0, res);
				continue;
			}

			selected = 1;

		} else if ((pid & 0xff00) == 0x1100) {
			for (j = 0; j < 12; ++j) {
				tmp = 0;
				r = read_pci(0x40 + j * 8 + 0, &tmp, 1);
				if (r || (tmp & 0x01) == 0)
					continue;

				r = read_pci(0x40 + j * 8 + 4, &tmp, 2);
				if (r)
					continue;

				if (tmp != res)
					continue;

				printf("%s: tmp: %xh, res:  %xh\n", __func__,
				       tmp, res);

				selected = 1;
			}

		} else {
			continue;
		}

		if (selected)
			break;
	}

	if (selected) {
		//printf("%s: %d\n", __func__, __LINE__);
		printf("%s: %04lxh %04lxh\n", m_pci_path, vid, pid);
		//printf("%s: %d\n", __func__, __LINE__);
	}

	for (i = 0; i < n; i++)
		free(namelist[i]);

	free(namelist);

	//printf("%s: selected: %d\n", __func__, selected);

	if (!selected)
		return -1;

	return 0;
}

static int is_f81504(void)
{
	int len;
	int tmp;
	char buf[5];

	memset(buf, 0, sizeof(buf));

	for (len = 0; len < 4; ++len) {
		read_pci_byte(len, &buf[len]);
		//printf("%x", buf[len]);
	}

	//printf("\n");

	if ((buf[0] != 0x29) || (buf[1] != 0x1c) || (buf[3] != 0x11))
		return -1;

	switch (buf[2]) {
	case 0x04:
	case 0x08:
	case 0x12:
		return 0;
	}

	return -1;
}

static int is_f81504a(void)
{
	int len;
	int tmp;
	char buf[5];

	memset(buf, 0, sizeof(buf));

	for (len = 0; len < 4; ++len) {
		read_pci_byte(len, &buf[len]);
	}

	//printf("\n");

	if ((buf[0] != 0x29) || (buf[1] != 0x1c) || (buf[3] != 0x13))
		return -1;

	switch (buf[2]) {
	case 0x04:
	case 0x08:
	case 0x12:
		return 0;
	}

	return -1;
}

static int get_tty_idx_f81504(unsigned int io_addr)
{
	unsigned char buf[2];
	unsigned int i, j;
	int read_io;
	int len;

	for (i = 0; i < 12; ++i) {
		memset(buf, 0, sizeof(buf));

		for (j = 0; j < 2; ++j) {
			read_pci_byte(0x40 + 0x8 * i + 4 + 0, &buf[0]);
			read_pci_byte(0x40 + 0x8 * i + 4 + 1, &buf[1]);
		}

		read_io = (buf[1] << 8) | buf[0];

		if (io_addr == read_io)
			break;
	}

	if (i >= 12)
		return -1;

	return i;
}

static int get_tty_idx_f81504a(unsigned int io_addr)
{
	uint32_t res, idx;
	int r;

	printf("%s: io_addr: %04xh, res: %02xh\n", __func__, io_addr, res);

	r = read_pci(0x10, &res, sizeof(uint32_t));
	if (r)
		return r;

	if (res > io_addr || ((io_addr - res) % 8))
		return -1;

	return (io_addr - res) / 8;
}

static uint32_t read_resource(char *tty)
{
	char buf[128];
	int len, tty_idx, cur_idx;
	int fd;
	FILE *file;
	int io_addr = 0;
	char *pos, *pos_space;

	do {
		sprintf(buf, F81504_CONFIG_SYSFS_IOPORT, tty);

		fd = open(buf, O_RDONLY);
		if (fd < 0) {
			break;
		}

		memset(buf, 0, sizeof(buf));
		len = read(fd, buf, sizeof(buf));
		if (!len) {
			close(fd);
			break;
		}

		io_addr = strtol(buf, NULL, 16);
		if (io_addr == 0) {
			close(fd);
			break;
		}

		close(fd);
		return io_addr;
	} while (0);

	do {
		sprintf(buf, F81504_CONFIG_SYSFS_MEM_BASE, tty);

		fd = open(buf, O_RDONLY);
		if (fd < 0) {
			break;
		}

		memset(buf, 0, sizeof(buf));
		len = read(fd, buf, sizeof(buf));
		if (!len) {
			close(fd);
			break;
		}

		io_addr = strtol(buf, NULL, 16);
		if (io_addr == 0) {
			close(fd);
			break;
		}

		close(fd);
		return io_addr;
	} while (0);

	sprintf(buf, F81504_SERIAL_SYSFS_IOPORT);
	file = fopen(buf, "r");
	if (!file)
		return 0;

	tty_idx = strtol(&tty[4], NULL, 10);

	io_addr = 0;
	while (fgets(buf, sizeof(buf), file)) {
		if (buf[0] < '0' || buf[0] > '9')
			continue;

		cur_idx = strtol(buf, NULL, 10);

		if (tty_idx != cur_idx)
			continue;

		//printf("%s", buf);

		pos = strstr(buf, "port:");
		pos_space = strstr(&buf[pos - buf + 5], " ");
		*pos_space = 0;

		//printf("pos: %d\n", pos - buf);
		//printf("%s\n", &buf[pos - buf + 5]);

		io_addr = strtol(&buf[pos - buf + 5], NULL, 16);
		break;
	}

	fclose(file);
	return io_addr;
}

static int get_transceiver_type(int idx, transceiver_type *type)
{
	unsigned char buf[128];
	FILE *file;
	int i, data[4];

	if (idx >= 4)
		return -1;

	sprintf(buf, "%s/transceiver", m_pci_path);

	file = fopen(buf, "r");
	if (!file) {
		printf("open failed: %s\n", buf);
		return -1;
	}

	for (i = 0; i < 4; ++i) {
		fscanf(file, "%d\n", &data[i]);

		if (i == idx) {
			*type = data[i];
			break;
		}
	}

	fclose(file);
	return 0;
}

static int get_current_mode(char *tty, int idx, unsigned int *mode)
{
	unsigned char buf[128];
	FILE *file;
	int i, data[4];

	if (idx >= 4)
		return -1;

	sprintf(buf, "%s/gpio_ctrl", m_pci_path);

	file = fopen(buf, "r");
	if (!file) {
		printf("open failed: %s\n", buf);
		return -1;
	}

	for (i = 0; i < 4; ++i) {
		fscanf(file, "%d\n", &data[i]);
		//printf("%s: %d %d\n", __func__, i, data[i]);

		if (i == idx)
			*mode = data[i];
	}

	return 0;
}

static int check_transceiver_mode(transceiver_type type, int mode)
{
	const int *tb = m_nAcceptArray[type];
	int i;

	for (i = 0; tb[i] != -1; ++i) {
		if (mode == tb[i])
			return 0;
	}

	return -1;
}

static int change_gpio(char *tty, int idx, unsigned int mode)
{
	unsigned char buf[128];
	FILE *file;
	int len;

	if (idx >= 4)
		return -1;

	sprintf(buf, "%s/gpio_ctrl", m_pci_path);

	file = fopen(buf, "w");
	if (!file) {
		printf("open failed: %s\n", buf);
		return -1;
	}

	sprintf(buf, "%d %d\n", idx, mode);

	len = fprintf(file, buf, strlen(buf));
	if (len <= 0) {
		printf("set gpio failed: %d\n", len);
		return -1;
	}

	fclose(file);

	return 0;
}

void dummy_open(char *dev)
{
	uint8_t buf[64];
	FILE *file;

	sprintf(buf, "/dev/%s", dev);

	file = fopen(buf, "r");
	if (file)
		fclose(file);
}

int main(int argc, char *argv[])
{
	int length = 0;
	int i = 0, gpio_mode, devid, idx, is_fix;
	int ret = 0;
	int status;
	uint32_t io;
	struct serial_rs485 rs485conf;
	transceiver_type type;

	printf("VERSION: %s\n", VERSION);

	if (argc < 2) {
		usage(argv);
		return -1;
	}

	io = read_resource(argv[1]);
	printf("resource: %xh\n", io);

	ret = get_res_with_pcie(io);
	if (ret) {
		printf("%s: cannot get data\n", __func__);
		return 1;
	}

	printf("Target: %s\n", m_pci_path);

	if (!is_f81504a()) {
		printf("%s: 504a\n", __func__);
		idx = get_tty_idx_f81504a(io);
	} else if (!is_f81504()) {
		printf("%s: 504\n", __func__);
		idx = get_tty_idx_f81504(io);
	} else {
		fprintf(stderr, "Not F81504/504A series\n");
		return status;
	}

	printf("%s: %s resource: %x\n", __func__, argv[1], io);
	printf("%s: %s idx: %d\n", __func__, argv[1], idx);

	if (io == 0) {
		printf("%s: resource err %x\n", __func__, io);
		return -1;
	}

	dummy_open(argv[1]);

	status = get_transceiver_type(idx, &type);
	if (status) {
		printf("Cant get transeciver type\n");
		return status;
	}

	if (type == transceiver_type_f81433) {
		printf("Transceiver: F81433\n");
	} else if (type == transceiver_type_f81439) {
		printf("Transceiver: F81439\n");
	} else if (type == transceiver_type_f81435) {
		printf("Transceiver: F81435\n");
	} else {
		printf("Unknown\n");
		return 1;
	}

	status = get_current_mode(argv[1], idx, &gpio_mode);
	if (status) {
		printf("Cant get gpio mode\n");
		return status;
	}

	if (gpio_mode == 0xff) {
		// fix mode
		printf("current mode is fix mode\n");
		return 1;
	}

	printf("current mode: %d\n", gpio_mode);

	if (argc <= 2)
		return 0;

	gpio_mode = strtol(argv[2], NULL, 0);

	status = check_transceiver_mode(type, gpio_mode);
	if (status) {
		printf("%s not support mode: %d\n", m_nNameArray[type],
		       gpio_mode);
		return 1;
	}

	status = change_gpio(argv[1], idx, gpio_mode);
	if (status) {
		printf("change gpio failed\n");
		return 1;
	}

	printf("set mode to: %d\n", gpio_mode);

	return 0;
}
